const std = @import("std");
const fs = std.fs;
const io = std.io;
const os = std.os;
const mem = std.mem;
const Options = @import("options.zig").Options;
const program_name = @import("options.zig").program_name;
const ps = fs.path.sep_str;
pub const dict_file = "en.txt";

const case_mask: u8 = 0x20;
var mem_buf: [0x10000]u8 = undefined;

pub fn main() !void {
    var fixed_alloc = std.heap.FixedBufferAllocator.init(&mem_buf);
    const allocator = &fixed_alloc.allocator;
    const args = try std.process.argsAlloc(allocator);
    defer std.process.argsFree(allocator, args);
    const std_out = io.getStdOut();
    const stream = std_out.outStream();
    var buffered = io.bufferedOutStream(stream);
    var buf_stream = buffered.outStream();
    const is_tty = std_out.isTty();

    const options = Options.parse(
        allocator,
        args[1..],
        .{ .add_newline = is_tty }
    ) catch |err| {
        return switch (err) {
            error.Help => {
                _ = try stream.write(Options.help_text);
            },
            else => err,
        };
    };
    defer options.deinit();

    const map = try mapFile(allocator, options);
    defer os.munmap(map);

    var seed_bytes: [@sizeOf(u64)]u8 = undefined;
    try std.crypto.randomBytes(&seed_bytes);
    const seed = mem.readIntNative(u64, &seed_bytes);
    var prng = std.rand.DefaultCsprng.init(seed);

    try writePassphrase(&buf_stream, &prng, map, options);

    try buffered.flush();
}

// zig fmt: off
pub fn mapFile(
    allocator: *mem.Allocator,
    options: Options
) !@TypeOf(os.mmap).ReturnType.Payload {
// zig fmt: on
    const default_dictionary =
        if (std.builtin.mode == .Debug)
            "share" ++ ps ++ dict_file
        else
            ".." ++ ps ++ "share" ++ ps ++ program_name ++ ps ++ dict_file;

    if (std.builtin.mode == .Debug) {
        const file = try fs.cwd().openFile(
            options.dictionary orelse default_dictionary,
            .{ .read = true }
        );
        defer file.close();
        const stat = try file.stat();
        return try os.mmap(null, stat.size, os.PROT_READ, os.MAP_PRIVATE, file.handle, 0);
    }

    var buf: [fs.MAX_PATH_BYTES]u8 = undefined;
    const file_path = options.dictionary orelse try fs.path.join(
        allocator,
        &[_][]const u8{
            try fs.selfExeDirPath(&buf), default_dictionary,
        }
    );
    defer allocator.free(file_path);
    const file = try fs.cwd().openFile(file_path, .{ .read = true });
    defer file.close();
    const stat = try file.stat();
    return try os.mmap(null, stat.size, os.PROT_READ, os.MAP_PRIVATE, file.handle, 0);
}

// zig fmt: off
pub fn writePassphrase(
    stream: var,
    prng: var,
    words: var,
    options: Options
) !void {
// zig fmt: on
    try writeRandomWords(stream, prng, words, options);

    if (options.add_symbol)
        try writeRandomSymbol(stream, prng);

    if (options.add_digit)
        try writeRandomDigit(stream, prng);

    if (options.add_newline)
        try stream.writeByte('\n');
}

// zig fmt: off
pub fn writeRandomWords(
    stream: var,
    prng: var,
    words: var,
    options: Options
) !void {
// zig fmt: on
    var rem = options.words;

    while (rem != 0) : (rem -= 1) {
        const word = try findWordSlice(words, prng);

        if (options.capitalize) {
            try stream.writeByte(word[0] & ~case_mask);
            _ = try stream.write(word[1..]);
        } else
            _ = try stream.write(word);

        if (rem != 1)
            if (options.separator) |sep|
                try stream.writeByte(sep);
    }
}

fn findWordSlice(words: var, prng: var) ![]const u8 {
    const i = prng.random.uintLessThan(usize, words.len);
    const e = i + (mem.indexOfScalar(u8, words[i..], '\n') orelse
        return error.NewlineNotFound);
    var b = e - 1;

    while (b != 0) : (b -= 1) {
        if (words[b] == '\n') {
            b += 1;
            break;
        }
    }

    return words[b..e];
}

fn writeRandomSymbol(stream: var, prng: var) !void {
    const symbols = [32]u8{
        '!', '"', '#', '$', '%', '&', '\'', '(',
        ')', '*', '+', ',', '-', '.', '/', ':',
        ';', '<', '=', '>', '?', '@', '[', '\\',
        ']', '^', '_', '`', '{', '|', '}', '~',
    };
    const i = prng.random.uintLessThan(usize, symbols.len);

    try stream.writeByte(symbols[i]);
}

fn writeRandomDigit(stream: var, prng: var) !void {
    try stream.writeByte(prng.random.intRangeAtMost(u8, '0', '9'));
}
