const std = @import("std");
const mem = std.mem;
const fs = std.fs;
pub const program_name = "passphrase";

pub const Options = struct {
    allocator: *mem.Allocator,
    dictionary: ?[]const u8,
    separator: ?u8,
    words: u8,
    capitalize: bool,
    add_symbol: bool,
    add_digit: bool,
    add_newline: bool,
    const Self = @This();

    pub fn init(allocator: *mem.Allocator, values: var) Self {
        const ArgType = @TypeOf(values);
        return Self{
            .allocator = allocator,
            .dictionary = null,
            .separator = if (@hasField(ArgType, "separator")) values.separator else '-',
            .words = if (@hasField(ArgType, "words")) values.words else 4,
            .capitalize = if (@hasField(ArgType, "capitalize")) values.capitalize else false,
            .add_symbol = if (@hasField(ArgType, "add_symbol")) values.add_symbol else false,
            .add_digit = if (@hasField(ArgType, "add_digit")) values.add_digit else false,
            .add_newline = if (@hasField(ArgType, "add_newline")) values.add_newline else false,
        };
    }

    pub fn deinit(self: Self) void {
        if (self.dictionary) |d|
            self.allocator.free(d);
    }

    pub fn parse(allocator: *mem.Allocator, args: [][]const u8, values: var) !Self {
        var options = Self.init(allocator, values);
        var i: usize = 0;

        while (i != args.len) : (i += 1) {
            const arg = args[i];

            if (arg.len < 2 or '-' != arg[0])
                return error.InvalidOption;

            for (arg[1..]) |b| {
                switch (b) {
                    'c' => options.capitalize = true,
                    'x' => options.add_symbol = true,
                    'd' => options.add_digit = true,
                    else => {
                        if ('h' == b or mem.eql(u8, arg, "--help"))
                            return error.Help;

                        i += 1;

                        if (args.len <= i)
                            return error.InvalidOption;

                        switch (b) {
                            'w' => options.words = try std.fmt.parseInt(u8, args[i], 10),
                            's' => {
                                const sep = args[i];

                                if (1 < sep.len)
                                    return error.InvalidSeparator;

                                options.separator = if (0 == sep.len) null else sep[0];
                            },
                            'f' => {
                                options.dictionary = try mem.dupe(allocator, u8, args[i]);
                            },
                            else => return error.InvalidOption,
                        }
                    },
                }
            }
        }

        return options;
    }

    pub const help_text = "Usage: " ++ program_name ++ " [options]\n" ++
        \\
        \\Options:
        \\  -w num    Generate a passphrase with num words [default: 4]
        \\  -s sep    Put a sep character (or '' for none) between the words [default: -]
        \\  -c        Capitalize each word
        \\  -x        Put a random symbol at the end
        \\  -d        Put a random digit at the end
        \\  -f file   Use file as dictionary (one word per line)
        \\  -h --help Show help and quit
        \\
    ;
};
