const std = @import("std");
const fs = std.fs;
const io = std.io;
const os = std.os;
const mem = std.mem;
const testing = std.testing;
const allocator = testing.allocator;
const expect = testing.expect;
const expectEqual = testing.expectEqual;
const expectError = testing.expectError;
const pp = @import("main.zig");
const Options = @import("options.zig").Options;

const words =
    \\sitjestillingi
    \\pubertetsalder
    \\knivstikkar
    \\profesjonist
    \\matreiddum
    \\randomisera
    \\aksleleddi
    \\lunarisk
    \\hostemikstur
    \\folklorist
    \\familiefest
    \\velkometale
    \\putleteare
    \\samletal
    \\bankgruppe
    \\hypokonder
    \\terskelfjord
    \\illiberalitet
    \\slagbord
    \\utklass
;

test "separator" {
    var args = [_][]const u8{"-s", "_"};
    const opts = try Options.parse(allocator, args[0..], .{});
    if (opts.separator) |sep| {
        expectEqual(sep, '_');
    } else
        unreachable;
}

test "invalid separator" {
    var args = [_][]const u8{"-s", ":-)"};
    expectError(error.InvalidSeparator, Options.parse(allocator, args[0..], .{}));
}

test "empty separator" {
    var args = [_][]const u8{"-s", ""};
    const opts = try Options.parse(allocator, args[0..], .{});
    expectEqual(opts.separator, null);
}

test "number of words" {
    var out_buf: [0x1000]u8 = undefined;
    var stream = io.fixedBufferStream(out_buf[0..]);
    var prng = std.rand.DefaultPrng.init(0);
    var opts = Options.init(allocator, .{});
    opts.words = 20;
    try pp.writeRandomWords(stream.outStream(), &prng, words, opts);
    var wc: u8 = 1;
    for (stream.getWritten()) |b| {
        if ('-' == b)
            wc += 1;
    }
    expectEqual(opts.words, wc);
}

test "add symbol" {
    var out_buf: [0x1000]u8 = undefined;
    var stream = io.fixedBufferStream(out_buf[0..]);
    var prng = std.rand.DefaultPrng.init(0);
    var opts = Options.init(allocator, .{});
    opts.words = 1;
    opts.add_symbol = true;
    const words2 = "apa\nvad\nlol\n";
    try pp.writePassphrase(stream.outStream(), &prng, words2, opts);
    const phrase = stream.getWritten();
    expectEqual(phrase.len, 4);
    const lb = phrase[3];
    expect(isSymbol(lb));
}

fn isSymbol(char: u8) bool {
    return ('!' <= char and char <= '/')
        or (':' <= char and char <= '@')
        or ('[' <= char and char <= '`')
        or ('{' <= char and char <= '~');
}

test "add digit" {
    var out_buf: [0x1000]u8 = undefined;
    var stream = io.fixedBufferStream(out_buf[0..]);
    var prng = std.rand.DefaultPrng.init(0);
    var opts = Options.init(allocator, .{});
    opts.words = 1;
    opts.add_digit = true;
    const words2 = "apa\nvad\nlol\n";
    try pp.writePassphrase(stream.outStream(), &prng, words2, opts);
    const phrase = stream.getWritten();
    expectEqual(phrase.len, 4);
    const lb = phrase[3];
    expect('0' <= lb and lb <= '9');
}

test "capitalize" {
    var out_buf: [0x1000]u8 = undefined;
    var stream = io.fixedBufferStream(out_buf[0..]);
    var prng = std.rand.DefaultPrng.init(0);
    var opts = Options.init(allocator, .{});
    opts.words = 3;
    opts.capitalize = true;
    try pp.writePassphrase(stream.outStream(), &prng, words, opts);
    var cc: u8 = 0;
    for (stream.getWritten()) |b| {
        if ('A' <= b and b <= 'Z')
            cc += 1;
    }
    expectEqual(cc, 3);
}
